#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Simple module for downloading data with PycURL"""

import gzip
import pycurl


def url_download(urldata, localpath, compress=False, follow=True):
    """
    Helper function for downloading remote data files with PycURL.

    - urldata: string(s), URL or list of URLs
    - localpath: path of the filename to which output will be written
    - compress: output compression
    """
    urls = urldata
    if isinstance(urldata, str):
        urls = [urldata]

    openfunc = open
    if compress is True:
        openfunc = gzip.open

    with openfunc(localpath, 'wb') as out:
        for url in urls:
            c = pycurl.Curl()
            c.setopt(c.URL, url)
            c.setopt(c.WRITEDATA, out)
            if follow:
                c.setopt(c.FOLLOWLOCATION, True)
            c.perform()
            c.close()



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Custom Genome database implementation for *Apis mellifera* OGS 1.0."""

from __future__ import print_function
import filecmp
import gzip
import re
import subprocess
import sys
import genhub


class Am10DB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(Am10DB, self).__init__(label, conf, workdir)
        assert self.config['source'] == 'am10'
        self.specbase = ('http://hymenopteragenome.org/drupal/sites/'
                         'hymenopteragenome.org.beebase/files/data')

    def __repr__(self):
        return 'OGS1.0'

    @property
    def gdnaurl(self):
        return '%s/%s' % (self.specbase, self.gdnafilename)

    @property
    def gff3url(self):
        return '%s/%s' % (self.specbase, self.gff3filename)

    @property
    def proturl(self):
        return '%s/%s' % (self.specbase, self.protfilename)

    def format_gdna(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            deflinematch = re.search('>gnl\|[^\|]+\|(\S+)', line)
            if deflinematch:
                seqid = deflinematch.group(1)
                line = line.replace('>', '>%s ' % seqid)
            print(line, end='', file=outstream)

    def format_prot(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            # No processing required currently.
            # If any is ever needed, do it here.
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        cmds = list()
        cmds.append('gunzip -c %s' % self.gff3path)
        cmds.append('genhub-glean-to-gff3.py')
        cmds.append('tidygff3')
        cmds.append('genhub-format-gff3.py --source am10 -')
        cmds.append('seq-reg.py - %s' % self.gdnafile)
        cmds.append('gt gff3 -sort -tidy -o %s -force' % self.gff3file)

        commands = ' | '.join(cmds)
        if debug:  # pragma: no cover
            print('DEBUG: running command: %s' % commands, file=logstream)
        proc = subprocess.Popen(commands, shell=True, universal_newlines=True,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               'illegal uppercase attribute "Shift"' not in line and \
               'has the wrong phase' not in line and \
               line != '':
                print(line, file=logstream)
        assert proc.returncode == 0, \
            'annot cleanup command failed: %s' % commands

    def gff3_protids(self, instream):
        for line in instream:
            if '\tmRNA\t' not in line:
                continue
            namematch = re.search('Name=([^;\n]+)', line)
            assert namematch, 'cannot parse mRNA name: ' + line
            yield namematch.group(1)

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse gene and iLocus IDs: %s' % attrs
                geneid = idmatch.group(1)
                ilocusid = idmatch.group(2)
                gene2loci[geneid] = ilocusid
            elif feattype == 'mRNA':
                pattern = 'Parent=([^;\n]+);.*Name=([^;\n]+)'
                idmatch = re.search(pattern, attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs
                protid = idmatch.group(2)
                geneid = idmatch.group(1)
                locusid = gene2loci[geneid]
                locusname = locusid2name[locusid]
                yield protid, locusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_gdna_format():
    """Am10 gDNA formatting"""
    conf = genhub.test_registry.genome('Am10')
    amel_db = Am10DB('Am10', conf, workdir='testdata/demo-workdir')
    amel_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Am10/Am10.gdna.fa'
    testoutfile = 'testdata/fasta/am10-gdna-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Am10 gDNA formatting failed'

    assert repr(amel_db) == 'OGS1.0'
    url = ('http://hymenopteragenome.org/drupal/sites/'
           'hymenopteragenome.org.beebase/files/data/Amel_2.0_scaffolds.fa.gz')
    assert amel_db.gdnaurl == url, amel_db.gdnaurl
    url = ('http://hymenopteragenome.org/drupal/sites/'
           'hymenopteragenome.org.beebase/files/data/amel_OGSv1.0.gff.gz')
    assert amel_db.gff3url == url, amel_db.gff3url
    url = ('http://hymenopteragenome.org/drupal/sites/'
           'hymenopteragenome.org.beebase/files/data/amel_OGSv1.0_pep.fa.gz')
    assert amel_db.proturl == url, amel_db.proturl


def test_annotation_am10():
    """Am10 annotation formatting"""
    conf = genhub.test_registry.genome('Am10')
    amel_db = Am10DB('Am10', conf, workdir='testdata/demo-workdir')
    amel_db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Am10/Am10.gff3'
    testfile = 'testdata/gff3/am10-format.gff3'
    assert filecmp.cmp(outfile, testfile), 'Am10 annotation formatting failed'


def test_proteins_am10():
    """Am10 protein formatting"""
    conf = genhub.test_registry.genome('Am10')
    amel_db = Am10DB('Am10', conf, workdir='testdata/demo-workdir')
    amel_db.preprocess_prot(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Am10/Am10.all.prot.fa'
    testoutfile = 'testdata/fasta/am10-prot-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Am10 protein formatting failed'


def test_protids():
    """Am10: extract protein IDs from GFF3"""
    conf = genhub.test_registry.genome('Am10')
    db = Am10DB('Am10', conf)
    protids = ['GB18127-PA', 'GB12533-PA', 'GB12334-PA', 'GB13832-PA',
               'GB10374-PA', 'GB11673-PA', 'GB17405-PA', 'GB15709-PA',
               'GB13454-PA', 'GB16202-PA', 'GB15597-PA', 'GB13411-PA',
               'GB18700-PA', 'GB10338-PA', 'GB12987-PA', 'GB19188-PA',
               'GB18764-PA', 'GB10796-PA', 'GB12389-PA', 'GB17508-PA',
               'GB10290-PA', 'GB11017-PA', 'GB15828-PA', 'GB12008-PA',
               'GB15518-PA', 'GB14106-PA', 'GB10878-PA', 'GB17802-PA',
               'GB17122-PA', 'GB18635-PA']
    infile = 'testdata/gff3/am10-format.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """Am10: extract protein-->iLocus mapping from GFF3"""
    conf = genhub.test_registry.genome('Am10')
    db = Am10DB('Am10', conf)
    mapping = {'GB18571-PA': 'Am10ILC-00345', 'GB13022-PA': 'Am10ILC-00347'}
    infile = 'testdata/gff3/am10-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import argparse
import re
import sys

parser = argparse.ArgumentParser()
parser.add_argument('species', help='4-letter species label')
parser.add_argument('gff3', type=argparse.FileType('r'), default=sys.stdin)
args = parser.parse_args()

for line in args.gff3:
    liilmatch = re.search('liil=(\d+)', line)
    riilmatch = re.search('riil=(\d+)', line)
    namematch = re.search('Name=([^;\n]+)', line)
    if not liilmatch or not riilmatch:
        continue

    lname = namematch.group(1)
    liil = liilmatch.group(1)
    riil = riilmatch.group(1)
    fields = '\t'.join([args.species, lname, liil, riil])
    print(fields)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import re
import sys

for line in sys.stdin:
    fields = line.split('\t')
    assert len(fields) == 9
    ftype = fields[2]
    assert ftype in ['mRNA', 'CDS']

    attrstring = fields[8]
    protmatch = re.search('GenePrediction (\S+)', attrstring)
    assert protmatch, line
    protid = protmatch.group(1)
    featid = protid
    if ftype == 'CDS':
        featid += '-CDS'
    attrs = list()
    attrs.append('ID=%s' % featid)
    if ftype == 'CDS':
        attrs.append('Parent=%s' % protid)
    attrs.append('Name=%s' % protid)
    fields[8] = ';'.join(attrs)

    newentry = '\t'.join(fields)
    print(newentry)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import filecmp
import sys
import genhub


def ids(db, logstream=sys.stderr):  # pragma: no cover
    """
    Retrieve protein IDs/accessions from the genome annotation.

    The `db` variable, a `GenomeDB` object, must implement a `gff3_protids`
    method for this retrieval.
    """
    if logstream is not None:
        logmsg = '[GenHub: %s] retrieving protein IDs' % db.config['species']
        print(logmsg, file=logstream)

    specdir = '%s/%s' % (db.workdir, db.label)
    infile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    outfile = '%s/%s.protids.txt' % (specdir, db.label)
    with open(infile, 'r') as instream, open(outfile, 'w') as outstream:
        for protid in db.gff3_protids(instream):
            print(protid, file=outstream)


def sequences(db, logstream=sys.stderr):
    """Extract protein sequences."""
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'extracting protein sequences'
        print(logmsg, file=logstream)

    specdir = '%s/%s' % (db.workdir, db.label)
    idfile = '%s/%s.protids.txt' % (specdir, db.label)
    seqfile = '%s/%s.all.prot.fa' % (specdir, db.label)
    outfile = '%s/%s.prot.fa' % (specdir, db.label)
    with open(idfile, 'r') as idstream, \
            open(seqfile, 'r') as seqstream, \
            open(outfile, 'w') as outstream:
        for defline, seq in genhub.fasta.select(idstream, seqstream):
            defline = '>gnl|%s|%s' % (db.label, defline[1:])
            print(defline, file=outstream)
            genhub.fasta.format(seq, outstream=outstream)


def mapping(db, logstream=sys.stderr):  # pragma: no cover
    """
    Retrieve mapping of protein IDs to iLocus IDs.

    The `db` variable, a `GenomeDB` object, must implement a `protein_mapping`
    method for this retrieval.
    """
    if logstream is not None:
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'parsing protein->iLocus mapping'
        print(logmsg, file=logstream)

    specdir = '%s/%s' % (db.workdir, db.label)
    infile = '%s/%s.iloci.gff3' % (specdir, db.label)
    outfile = '%s/%s.protein2ilocus.txt' % (specdir, db.label)
    with open(infile, 'r') as instream, open(outfile, 'w') as outstream:
        for protid, ilocusid in db.protein_mapping(instream):
            print(protid, ilocusid, sep='\t', file=outstream)


# -----------------------------------------------------------------------------
# Driver function
# -----------------------------------------------------------------------------

def prepare(db, logstream=sys.stderr):  # pragma: no cover
    ids(db, logstream=logstream)
    sequences(db, logstream=logstream)
    mapping(db, logstream=logstream)


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_protein_sequence():
    """Select protein sequences"""
    config = genhub.test_registry.genome('Scer')
    d = genhub.refseq.RefSeqDB('Scer', config, workdir='testdata/demo-workdir')
    sequences(d, logstream=None)

    outfile = 'testdata/demo-workdir/Scer/Scer.prot.fa'
    testfile = 'testdata/fasta/scer-few-prots.fa'
    assert filecmp.cmp(outfile, testfile), 'Protein sequence selection failed'



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import filecmp
import re
import subprocess
import sys
import genhub


def cds_sequences(db, logstream=sys.stderr):
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'extracting coding sequences'
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)

    gff3infile = '%s/%s.gff3' % (specdir, db.label)
    fastainfile = '%s/%s.gdna.fa' % (specdir, db.label)
    outfile = '%s/%s.all.cds.fa' % (specdir, db.label)
    command = 'xtractore --type=CDS --outfile=%s ' % outfile
    command += '%s %s' % (gff3infile, fastainfile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)

    gff3infile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    fastainfile = '%s/%s.gdna.fa' % (specdir, db.label)
    outfile = '%s/%s.cds.fa' % (specdir, db.label)
    command = 'xtractore --type=CDS --outfile=%s ' % outfile
    command += '%s %s' % (gff3infile, fastainfile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)


def exon_sequences(db, logstream=sys.stderr):
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'extracting exon sequences'
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)

    gff3infile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    fastainfile = '%s/%s.gdna.fa' % (specdir, db.label)
    outfile = '%s/%s.exons.fa' % (specdir, db.label)
    command = 'xtractore --type=exon --outfile=%s ' % outfile
    command += '%s %s' % (gff3infile, fastainfile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)


def parse_intron_accessions(instream):
    moltypes = ['mRNA', 'tRNA', 'ncRNA', 'transcript', 'primary_transcript',
                'V_gene_segment', 'D_gene_segment', 'J_gene_segment',
                'C_gene_segment']
    id_to_accession = dict()
    for line in instream:
        line = line.rstrip()
        idmatch = re.search('ID=([^;\n]+)', line)
        accmatch = re.search('accession=([^;\n]+)', line)
        if idmatch and accmatch:
            molid = idmatch.group(1)
            accession = accmatch.group(1)
            id_to_accession[molid] = accession

        if '\tintron\t' in line:
            parentid = re.search('Parent=([^;\n]+)', line).group(1)
            assert ',' not in parentid, parentid
            accession = id_to_accession[parentid]
            line += ';accession=%s' % accession

        yield(line)


def intron_sequences(db, logstream=sys.stderr):
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'extracting intron sequences'
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)

    infile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    outfile = '%s/%s.with-introns.gff3' % (specdir, db.label)
    command = 'canon-gff3 --outfile=%s %s' % (outfile, infile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)

    infile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    outfile = '%s/%s.with-introns.gff3' % (specdir, db.label)
    with open(infile, 'r') as instream, open(outfile, 'w') as outstream:
        for line in parse_intron_accessions(instream):
            print(line, file=outstream)

    gff3infile = '%s/%s.with-introns.gff3' % (specdir, db.label)
    fastainfile = '%s/%s.gdna.fa' % (specdir, db.label)
    outfile = '%s/%s.introns.fa' % (specdir, db.label)
    command = 'xtractore --type=intron --outfile=%s ' % outfile
    command += '%s %s' % (gff3infile, fastainfile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)


# -----------------------------------------------------------------------------
# Driver function
# -----------------------------------------------------------------------------


def prepare(db, logstream=sys.stderr):  # pragma: no cover
    exon_sequences(db, logstream=logstream)
    intron_sequences(db, logstream=logstream)
    cds_sequences(db, logstream=logstream)


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_coding_sequences():
    """Extract coding sequences"""
    config = genhub.test_registry.genome('Atha')
    db = genhub.refseq.RefSeqDB('Atha', config,
                                workdir='testdata/demo-workdir')
    cds_sequences(db, logstream=None)

    outfile = 'testdata/demo-workdir/Atha/Atha.cds.fa'
    testfile = 'testdata/fasta/atha-cds.fa'
    assert filecmp.cmp(outfile, testfile), 'coding sequence extraction failed'


def test_exon_sequences():
    """Extract exon sequences"""
    config = genhub.test_registry.genome('Atha')
    db = genhub.refseq.RefSeqDB('Atha', config,
                                workdir='testdata/demo-workdir')
    exon_sequences(db, logstream=None)

    outfile = 'testdata/demo-workdir/Atha/Atha.exons.fa'
    testfile = 'testdata/fasta/atha-exons.fa'
    assert filecmp.cmp(outfile, testfile), 'exon sequence extraction failed'


def test_intron_sequences():
    """Extract intron sequences"""
    config = genhub.test_registry.genome('Atha')
    db = genhub.refseq.RefSeqDB('Atha', config,
                                workdir='testdata/demo-workdir')
    intron_sequences(db, logstream=None)

    outfile = 'testdata/demo-workdir/Atha/Atha.introns.fa'
    testfile = 'testdata/fasta/atha-introns.fa'
    assert filecmp.cmp(outfile, testfile), 'intron sequence extraction failed'



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import re
import sys


def process_gff3(fp):
    trnas = {}
    for line in fp:
        line = line.rstrip()
        fields = line.split("\t")
        if len(fields) != 9:
            yield line
            continue
        if fields[2] == "gene" and "gene=tRNA" in fields[8]:
            geneid = re.search("Name=tRNA:([^;\n]+)", fields[8]).group(1)
            fields[8] = "ID=gene.trna.%s;%s" % (geneid, fields[8])
            yield "\t".join(fields)
            continue
        elif fields[2] == "tRNA":
            geneid = re.search("Note=tRNA:(.+?)-RA", fields[8]).group(1)
            fields[8] = "Parent=gene.trna.%s;%s" % (geneid, fields[8])
            if geneid in trnas:
                otherfields = trnas[geneid]
                fields[3] = min(fields[3], otherfields[3])
                fields[4] = max(fields[4], otherfields[4])
            trnas[geneid] = fields
            continue
        yield line

    for geneid in trnas:
        yield "\t".join(trnas[geneid])

if __name__ == "__main__":
    for line in process_gff3(sys.stdin):
        print(line)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Genome database implementation for *Polites dominula* genome data."""

from __future__ import print_function
import filecmp
import gzip
import re
import subprocess
import sys
import genhub


class PdomDB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(PdomDB, self).__init__(label, conf, workdir)
        assert self.config['source'] == 'pdom'
        self.specbase = 'http://de.iplantcollaborative.org/dl/d'

    def __repr__(self):
        return 'PdomDataStore'

    @property
    def gdnaurl(self):
        prefix = '53B7319E-3201-4087-9607-2D541FF34DD0'
        return '%s/%s/%s' % (self.specbase, prefix, self.gdnafilename)

    @property
    def gff3url(self):
        prefix = 'E4944CBB-7DE4-4CA1-A889-3D2A5D2E8696'
        return '%s/%s/%s' % (self.specbase, prefix, self.gff3filename)

    @property
    def proturl(self):
        prefix = 'ACD29139-6619-48DF-A9F2-F75CA382E248'
        return '%s/%s/%s' % (self.specbase, prefix, self.protfilename)

    def format_gdna(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            # No processing required currently.
            # If any is ever needed, do it here.
            print(line, end='', file=outstream)

    def format_prot(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            # No processing required currently.
            # If any is ever needed, do it here.
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        command = ['genhub-format-gff3.py', '--source', 'pdom', '--outfile',
                   self.gff3file, self.gff3path]
        subprocess.check_call(command)

    def gff3_protids(self, instream):
        for line in instream:
            if '\tmRNA\t' not in line:
                continue
            namematch = re.search('Name=([^;\n]+)', line)
            assert namematch, 'cannot parse mRNA name: ' + line
            yield namematch.group(1)

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse gene and iLocus IDs: %s' % attrs
                geneid = idmatch.group(1)
                ilocusid = idmatch.group(2)
                gene2loci[geneid] = ilocusid
            elif feattype == 'mRNA':
                pattern = 'Parent=([^;\n]+);.*Name=([^;\n]+)'
                idmatch = re.search(pattern, attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs
                protid = idmatch.group(2)
                geneid = idmatch.group(1)
                locusid = gene2loci[geneid]
                locusname = locusid2name[locusid]
                yield protid, locusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------

def test_download():
    """PdomDataStore download"""
    config = genhub.test_registry.genome('Pdom')
    pdom_db = PdomDB('Pdom', config)

    assert pdom_db.gdnaurl == ('http://de.iplantcollaborative.org/dl/d/'
                               '53B7319E-3201-4087-9607-2D541FF34DD0/'
                               'pdom-scaffolds-unmasked-r1.2.fa.gz')
    assert pdom_db.gff3url == ('http://de.iplantcollaborative.org/dl/d/'
                               'E4944CBB-7DE4-4CA1-A889-3D2A5D2E8696/'
                               'pdom-annot-r1.2.gff3')
    assert pdom_db.proturl == ('http://de.iplantcollaborative.org/dl/d/'
                               'ACD29139-6619-48DF-A9F2-F75CA382E248/'
                               'pdom-annot-r1.2-proteins.fa')
    assert '%r' % pdom_db == 'PdomDataStore'


def test_format():
    """Pdom formatting task"""
    config = genhub.test_registry_supp.genome('Pdom')
    pdom_db = PdomDB('Pdom', config, workdir='testdata/demo-workdir')

    pdom_db.preprocess_gdna(logstream=None)
    pdom_db.preprocess_gff3(logstream=None)
    pdom_db.preprocess_prot(logstream=None)


def test_protids():
    """Pdom: extract protein IDs from GFF3"""
    config = genhub.test_registry.genome('Pdom')
    db = PdomDB('Pdom', config)

    protids = ['PdomMRNAr1.2-08518.1', 'PdomMRNAr1.2-11420.1',
               'PdomMRNAr1.2-08519.1']
    infile = 'testdata/gff3/pdom-266.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """Pdom: extract protein-->iLocus mapping from GFF3"""
    config = genhub.test_registry.genome('Pdom')
    db = PdomDB('Pdom', config)

    mapping = {'PdomMRNAr1.2-08518.1': 'PdomILC-18235',
               'PdomMRNAr1.2-11420.1': 'PdomILC-18237',
               'PdomMRNAr1.2-08519.1': 'PdomILC-18238'}
    infile = 'testdata/gff3/pdom-266-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""
Retrieve and format data from CRG.

GenomeDB implementation for two data sets residing in an unofficial
distribution site at the Centre de Regulacio Genomica.
"""

from __future__ import print_function
import re
import subprocess
import sys
import genhub


class CrgDB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(CrgDB, self).__init__(label, conf, workdir)
        assert self.config['source'] == 'crg'
        self.specbase = 'http://wasp.crg.eu'

    def __repr__(self):
        return 'CRG'

    @property
    def gdnaurl(self):
        return '%s/%s' % (self.specbase, self.gdnafilename)

    @property
    def gff3url(self):
        return '%s/%s' % (self.specbase, self.config['annotation'])

    @property
    def proturl(self):
        return '%s/%s' % (self.specbase, self.protfilename)

    @property
    def gff3filename(self):
        return '%s.gz' % self.config['annotation']

    def format_gdna(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            if line.startswith('>'):
                line = line.replace('scaffold_', '%sScf_' % self.label)
                line = line.replace('scaffold', '%sScf_' % self.label)
            print(line, end='', file=outstream)

    def format_prot(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            # No processing required currently.
            # If any is ever needed, do it here.
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        cmds = list()
        cmds.append('gunzip -c %s' % self.gff3path)
        cmds.append("sed 's/	transcript	/	mRNA	/'")
        cmds.append("sed 's/scaffold_/%sScf_/'" % self.label)
        cmds.append("sed 's/scaffold/%sScf_/'" % self.label)
        cmds.append('genhub-format-gff3.py --source crg -')
        cmds.append('seq-reg.py - %s' % self.gdnafile)
        cmds.append('gt gff3 -sort -tidy -o %s -force' % self.gff3file)

        commands = ' | '.join(cmds)
        if debug:  # pragma: no cover
            print('DEBUG: running command: %s' % commands, file=logstream)
        proc = subprocess.Popen(commands, shell=True, universal_newlines=True,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               line != '':
                print(line, file=logstream)
        assert proc.returncode == 0, \
            'annot cleanup command failed: %s' % commands

    def gff3_protids(self, instream):
        protids = dict()
        for line in instream:
            if '\tCDS\t' not in line:
                continue
            idmatch = re.search('Target=(\S+)', line)
            assert idmatch, 'cannot parse protein_id: ' + line
            protid = idmatch.group(1)
            if protid not in protids:
                protids[protid] = True
                yield protid

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        mrna2gene = dict()
        proteins = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse gene and iLocus IDs: %s' % attrs
                geneid = idmatch.group(1)
                ilocusid = idmatch.group(2)
                gene2loci[geneid] = ilocusid
            elif feattype == 'mRNA':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs
                mrnaid = idmatch.group(1)
                geneid = idmatch.group(2)
                mrna2gene[mrnaid] = geneid
            elif feattype == 'CDS':
                idmatch = re.search('Parent=([^;\n]+).*Target=(\S+)', attrs)
                assert idmatch, \
                    'Unable to parse protein and mRNA IDs: %s' % attrs
                mrnaid = idmatch.group(1)
                proteinid = idmatch.group(2)
                if proteinid not in proteins:
                    geneid = mrna2gene[mrnaid]
                    ilocusid = gene2loci[geneid]
                    ilocusname = locusid2name[ilocusid]
                    proteins[proteinid] = mrnaid
                    yield proteinid, ilocusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_scaffolds():
    """CRG scaffolds download"""
    config = genhub.test_registry.genome('Dqua')
    testurl = 'http://wasp.crg.eu/DQUA.v01.fa.gz'
    testpath = './Dqua/DQUA.v01.fa.gz'
    dqua_db = CrgDB('Dqua', config)
    assert dqua_db.gdnaurl == testurl, \
        'scaffold URL mismatch\n%s\n%s' % (dqua_db.gdnaurl, testurl)
    assert dqua_db.gdnapath == testpath, \
        'scaffold path mismatch\n%s\n%s' % (dqua_db.gdnapath, testpath)
    assert '%r' % dqua_db == 'CRG'


def test_annot():
    """CRG annotation download"""
    config = genhub.test_registry.genome('Dqua')
    testurl = 'http://wasp.crg.eu/DQUA.v01.gff3'
    testpath = 'CRG/Dqua/DQUA.v01.gff3.gz'
    dqua_db = CrgDB('Dqua', config, workdir='CRG')
    assert dqua_db.gff3url == testurl, \
        'annotation URL mismatch\n%s\n%s' % (dqua_db.gff3url, testurl)
    assert dqua_db.gff3path == testpath, \
        'annotation path mismatch\n%s\n%s' % (dqua_db.gff3path, testpath)


def test_proteins():
    """CRG protein download"""
    config = genhub.test_registry.genome('Dqua')
    testurl = 'http://wasp.crg.eu/DQUA.v01.pep.fa.gz'
    testpath = '/opt/db/genhub/Dqua/DQUA.v01.pep.fa.gz'
    dqua_db = CrgDB('Dqua', config, workdir='/opt/db/genhub')
    assert dqua_db.proturl == testurl, \
        'protein URL mismatch\n%s\n%s' % (dqua_db.proturl, testurl)
    assert dqua_db.protpath == testpath, \
        'protein path mismatch\n%s\n%s' % (dqua_db.protpath, testpath)


def test_protids():
    """CRG: extract protein IDs from GFF3"""
    conf = genhub.test_registry.genome('Dqua')
    db = CrgDB('Dqua', conf)
    protids = ['DQUA011a006022P1', 'DQUA011a006023P1', 'DQUA011a006024P1']
    infile = 'testdata/gff3/dqua-275.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """CRG: extract protein-->iLocus mapping from GFF3"""
    conf = genhub.test_registry.genome('Dqua')
    db = CrgDB('Dqua', conf)
    mapping = {'DQUA011a006022P1': 'DquaILC-14465',
               'DQUA011a006023P1': 'DquaILC-14466',
               'DQUA011a006024P1': 'DquaILC-14467'}
    infile = 'testdata/gff3/dqua-275-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)


def test_format():
    """GenomeDB task drivers"""
    conf = genhub.test_registry_supp.genome('Pcan')
    pcan_db = CrgDB('Pcan', conf, workdir='testdata/demo-workdir')
    pcan_db.format(logstream=None)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import argparse
import importlib
import multiprocessing
import os
import subprocess
import sys
import genhub

buildcmds = ['list', 'download', 'format', 'prepare', 'stats', 'cluster',
             'cleanup']
sources = ['refseq', 'ncbi_flybase', 'beebase', 'crg', 'pdom', 'tair', 'am10']
dbtype = {'refseq': genhub.refseq.RefSeqDB,
          'ncbi_flybase': genhub.ncbi_flybase.FlyBaseDB,
          'beebase': genhub.beebase.BeeBaseDB,
          'crg': genhub.crg.CrgDB,
          'pdom': genhub.pdom.PdomDB,
          'tair': genhub.tair.TairDB,
          'am10': genhub.am10.Am10DB}


def getdb(label, config, args):
    assert 'source' in config
    assert config['source'] in sources
    constructor = dbtype[config['source']]
    db = constructor(label, config, workdir=args.workdir)
    return db


def list_configs(registry):
    print('===== Genome configurations =====')
    for label, config in registry.list_genomes():
        print(label, config['species'], sep='\t', end='')
        if 'common' in config:
            print('\t', config['common'], end='')
        print('')
    print('')

    print('===== Batch configurations =====')
    for label, batch in registry.list_batches():
        print(label, '\t', ','.join(batch))


def run_build(builddata):
    label, config, args = builddata
    db = getdb(label, config, args)

    if 'download' in args.task:
        db.download()
    if 'format' in args.task:
        db.format()
    if 'prepare' in args.task:
        genhub.iloci.prepare(db, ilcformat=args.format)
        genhub.proteins.prepare(db)
        genhub.mrnas.prepare(db)
        genhub.exons.prepare(db)
    if 'stats' in args.task:
        genhub.stats.compute(db)
    if 'cleanup' in args.task:
        db.cleanup(args.keep, args.fullclean)

    print('[GenHub: %s] build complete!' % db.config['species'],
          file=sys.stderr)


def cluster_proteins(dbs, np=1):
    print('[GenHub] aggregating representative proteins', file=sys.stderr)
    protmap = dict()
    with open('GenHub.prot.fa', 'w') as outstream:
        for db in dbs:
            protfile = '%s/%s.prot.fa' % (db.dbdir, db.label)
            with open(protfile, 'r') as instream:
                for line in instream:
                    print(line, end='', file=outstream)
            for protid, locid in db.get_prot_map():
                protmap[protid] = locid

    print('[GenHub] clustering representative proteins', file=sys.stderr)
    command = ('cd-hit -i GenHub.prot.fa -o GenHub.prot -T %d -d 0 '
               '-c 0.50 -s 0.65 -p 1 -n 3 -aL 0.75 -aS 0.85 -g 1' % np).split()
    subprocess.check_call(command)

    with open('GenHub.prot.clstr', 'r') as infile, \
            open('GenHub.hiloci.tsv', 'w') as outfile:
        for clusterid, clusterseqs in genhub.cdhit.parse_clusters(infile):
            iloci = [protmap[prot.accession] for prot in clusterseqs]
            species = set([prot.species for prot in clusterseqs])
            print(len(iloci), len(species), ','.join(iloci), ','.join(species),
                  sep='\t', file=outfile)


def get_parser():
    desc = 'Run the main GenHub build process.'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-v', '--version', action='version',
                        version='GenHub v%s' % genhub.__version__)
    parser.add_argument('-w', '--workdir', metavar='WD', default='./species',
                        help='working directory for data files; default is '
                        '"./species"')
    parser.add_argument('-c', '--cfgdir', default=None, metavar='DIR',
                        help='Directory (or comma-separated list of '
                        'directories) from which to load user-supplied genome '
                        'configuration files')
    parser.add_argument('-p', '--numprocs', metavar='P', type=int, default=1,
                        help='number of processors to use when processing '
                        'multiple genomes; default is 1')
    parser.add_argument('--keep', metavar='PTN', nargs='+',
                        help='keep files matching the specified pattern PTN '
                        'when running the `cleanup` build task')
    parser.add_argument('--fullclean', action='store_true',
                        help='when running the `cleanup` build task, delete '
                        'original (downloaded) data files as well as processed'
                        ' data files')
    parser.add_argument('-f', '--format', metavar='PFX', default='{}ILC-%05lu',
                        help='format for assigning serial labels to iLoci; '
                        'must include the placeholder {} for the species '
                        'label, as well as a printf-style placeholder for a '
                        'serial number; default is "{}ILC-%%05lu"')
    confargs = parser.add_mutually_exclusive_group()
    confargs.add_argument('-g', '--genome', default=None, metavar='LBL',
                          help='Label (or comma-separated set of labels) '
                          'specifying the genome(s) to process; use the'
                          '`list` task to show all available genomes')
    confargs.add_argument('-b', '--batch', default=None, metavar='LBL',
                          help='Label of a batch of genomes to process; use '
                          'the `list` task to show all available '
                          'batches')
    parser.add_argument('task', nargs='+', choices=buildcmds, metavar='task',
                        help='Build task(s) to execute; options include '
                        '"%s"' % '", "'.join(buildcmds))
    return parser


def main(args):
    registry = genhub.registry.Registry()
    if args.cfgdir:
        for cfgdirpath in args.cfgdir.split(','):
            registry.update(cfgdirpath)

    if 'list' in args.task:
        list_configs(registry)
        sys.exit(0)

    if args.genome:
        labels = args.genome.split(',')
        conf = registry.genomes(labels)
        assert conf is not None, 'unknown genome label(s) "%s"' % args.genome
    elif args.batch:
        conf = registry.batch(args.batch)
        assert conf is not None, 'unknown batch label "%s"' % args.batch
    else:
        message = ('must specify a genome or batch of genomes to process, '
                   'or `list` to show available genomes')
        raise ValueError(message)

    builds = list()
    for label in sorted(conf):
        builddata = (label, conf[label], args)
        builds.append(builddata)
    pool = multiprocessing.Pool(processes=args.numprocs)
    results = [pool.apply_async(run_build, args=(b,)) for b in builds]
    _ = [p.get() for p in results]

    if 'cluster' in args.task:
        dbs = [getdb(label, conf[label], args) for label in sorted(conf)]
        cluster_proteins(dbs, args.numprocs)

    print('[GenHub] all builds complete!', file=sys.stderr)


if __name__ == '__main__':
    main(get_parser().parse_args())



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""
Retrieve and format data from BeeBase.

GenomeDB implementation for BeeBase consortium data provisioned by
HymenopteraBase.
"""

from __future__ import print_function
import filecmp
import gzip
import re
import subprocess
import sys
import genhub


class BeeBaseDB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(BeeBaseDB, self).__init__(label, conf, workdir)
        assert self.config['source'] == 'beebase'
        self.specbase = ('http://hymenopteragenome.org/beebase/sites/'
                         'hymenopteragenome.org.beebase/files/data/'
                         'consortium_data')

    def __repr__(self):
        return 'BeeBase'

    @property
    def gdnaurl(self):
        return '%s/%s' % (self.specbase, self.gdnafilename)

    @property
    def gff3url(self):
        return '%s/%s' % (self.specbase, self.gff3filename)

    @property
    def proturl(self):
        return '%s/%s' % (self.specbase, self.protfilename)

    def format_gdna(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            if line.startswith('>scaffold'):
                line = line.replace('scaffold', '%sScf_' % self.label)
            elif line.startswith('>Group'):
                line = line.replace('Group', '%sGroup' % self.label)
            print(line, end='', file=outstream)

    def format_prot(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            if line.startswith('>gnl|'):
                deflinematch = re.search('>gnl\|[^\|]+\|(\S+)', line)
                assert deflinematch, line
                protid = deflinematch.group(1)
                line = line.replace('>', '>%s ' % protid)
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        cmds = list()
        cmds.append('gunzip -c %s' % self.gff3path)
        cmds.append('genhub-namedup.py')
        cmds.append("sed 's/scaffold/%sScf_/'" % self.label)
        cmds.append("sed 's/Group/%sGroup/'" % self.label)
        cmds.append('tidygff3')
        cmds.append('genhub-format-gff3.py --source beebase -')
        cmds.append('seq-reg.py - %s' % self.gdnafile)
        cmds.append('gt gff3 -sort -tidy -o %s -force' % self.gff3file)

        commands = ' | '.join(cmds)
        if debug:  # pragma: no cover
            print('DEBUG: running command: %s' % commands, file=logstream)
        proc = subprocess.Popen(commands, shell=True, universal_newlines=True,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               'illegal uppercase attribute "Shift"' not in line and \
               'has the wrong phase' not in line and \
               line != '':
                print(line, file=logstream)
        assert proc.returncode == 0, \
            'annot cleanup command failed: %s' % commands

    def gff3_protids(self, instream):
        for line in instream:
            if '\tmRNA\t' not in line:
                continue
            namematch = re.search('Name=([^;\n]+)', line)
            assert namematch, 'cannot parse mRNA name: ' + line
            protid = namematch.group(1).replace('-RA', '-PA')
            yield protid

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse gene and iLocus IDs: %s' % attrs
                geneid = idmatch.group(1)
                ilocusid = idmatch.group(2)
                gene2loci[geneid] = ilocusid
            elif feattype == 'mRNA':
                pattern = 'Parent=([^;\n]+);.*Name=([^;\n]+)'
                idmatch = re.search(pattern, attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs
                protid = idmatch.group(2).replace('-RA', '-PA')
                geneid = idmatch.group(1)
                locusid = gene2loci[geneid]
                locusname = locusid2name[locusid]
                yield protid, locusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_scaffolds_download():
    """BeeBase consortium scaffolds download"""
    config = genhub.test_registry.genome('Emex')
    testurl = ('http://hymenopteragenome.org/beebase/sites/'
               'hymenopteragenome.org.beebase/files/data/consortium_data/'
               'Eufriesea_mexicana.v1.0.fa.gz')
    testpath = './Emex/Eufriesea_mexicana.v1.0.fa.gz'
    emex_db = BeeBaseDB('Emex', config)
    assert emex_db.gdnaurl == testurl, \
        'scaffold URL mismatch\n%s\n%s' % (emex_db.gdnaurl, testurl)
    assert emex_db.gdnapath == testpath, \
        'scaffold path mismatch\n%s\n%s' % (emex_db.gdnapath, testpath)
    assert '%r' % emex_db == 'BeeBase'


def test_annot_download():
    """BeeBase consortium annotation download"""
    config = genhub.test_registry.genome('Dnov')
    testurl = ('http://hymenopteragenome.org/beebase/sites/'
               'hymenopteragenome.org.beebase/files/data/consortium_data/'
               'Dufourea_novaeangliae_v1.1.gff.gz')
    testpath = 'BeeBase/Dnov/Dufourea_novaeangliae_v1.1.gff.gz'
    dnov_db = BeeBaseDB('Dnov', config, workdir='BeeBase')
    assert dnov_db.gff3url == testurl, \
        'annotation URL mismatch\n%s\n%s' % (dnov_db.gff3url, testurl)
    assert dnov_db.gff3path == testpath, \
        'annotation path mismatch\n%s\n%s' % (dnov_db.gff3path, testpath)


def test_proteins_download():
    """BeeBase consortium protein download"""
    config = genhub.test_registry.genome('Hlab')
    testurl = ('http://hymenopteragenome.org/beebase/sites/'
               'hymenopteragenome.org.beebase/files/data/consortium_data/'
               'Habropoda_laboriosa_v1.2.pep.fa.gz')
    testpath = '/opt/db/genhub/Hlab/Habropoda_laboriosa_v1.2.pep.fa.gz'
    hlab_db = BeeBaseDB('Hlab', config, workdir='/opt/db/genhub')
    assert hlab_db.proturl == testurl, \
        'protein URL mismatch\n%s\n%s' % (hlab_db.proturl, testurl)
    assert hlab_db.protpath == testpath, \
        'protein path mismatch\n%s\n%s' % (hlab_db.protpath, testpath)


def test_gdna_format():
    """BeeBase gDNA formatting"""
    conf = genhub.test_registry.genome('Hlab')
    hlab_db = BeeBaseDB('Hlab', conf, workdir='testdata/demo-workdir')
    hlab_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Hlab/Hlab.gdna.fa'
    testoutfile = 'testdata/fasta/hlab-first-6-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Hlab gDNA formatting failed'

    conf = genhub.test_registry.genome('Am32')
    amel_db = BeeBaseDB('Am32', conf, workdir='testdata/demo-workdir')
    amel_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Am32/Am32.gdna.fa'
    testoutfile = 'testdata/fasta/am32-gdna-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Am32 gDNA formatting failed'


def test_annotation_beebase():
    """BeeBase annotation formatting"""
    conf = genhub.test_registry.genome('Hlab')
    hlab_db = BeeBaseDB('Hlab', conf, workdir='testdata/demo-workdir')
    hlab_db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Hlab/Hlab.gff3'
    testfile = 'testdata/gff3/beebase-format-hlab.gff3'
    assert filecmp.cmp(outfile, testfile), 'Hlab annotation formatting failed'


def test_proteins_beebase():
    """BeeBase protein formatting"""
    conf = genhub.test_registry.genome('Hlab')
    hlab_db = BeeBaseDB('Hlab', conf, workdir='testdata/demo-workdir')
    hlab_db.preprocess_prot(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Hlab/Hlab.all.prot.fa'
    testoutfile = 'testdata/fasta/hlab-first-20-prot-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Hlab protein formatting failed'

    conf = genhub.test_registry.genome('Am32')
    amel_db = BeeBaseDB('Am32', conf, workdir='testdata/demo-workdir')
    amel_db.preprocess_prot(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Am32/Am32.all.prot.fa'
    testoutfile = 'testdata/fasta/am32-prot-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Am32 protein formatting failed'


def test_protids():
    """BeeBase: extract protein IDs from GFF3"""
    conf = genhub.test_registry.genome('Hlab')
    db = BeeBaseDB('Hlab', conf)
    protids = ['Hlab050%d' % x for x in range(62, 75)]
    infile = 'testdata/gff3/hlab-238.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """BeeBase: extract protein-->iLocus mapping from GFF3"""
    conf = genhub.test_registry.genome('Hlab')
    db = BeeBaseDB('Hlab', conf)
    mapping = {'Hlab05074': 'HlabILC-653102', 'Hlab05071': 'HlabILC-653096',
               'Hlab05062': 'HlabILC-653079', 'Hlab05068': 'HlabILC-653090',
               'Hlab05072': 'HlabILC-653098', 'Hlab05070': 'HlabILC-653094',
               'Hlab05066': 'HlabILC-653086', 'Hlab05067': 'HlabILC-653088',
               'Hlab05069': 'HlabILC-653092', 'Hlab05073': 'HlabILC-653100',
               'Hlab05064': 'HlabILC-653083', 'Hlab05065': 'HlabILC-653085',
               'Hlab05063': 'HlabILC-653081'}
    infile = 'testdata/gff3/hlab-238-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import subprocess
import sys


def compute(db, logstream=sys.stderr):  # pragma: no cover
    if logstream is not None:
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'calculating feature statistics'
        print(logmsg, file=logstream)

    prefix = '%s/%s/%s' % (db.workdir, db.label, db.label)
    prefix3 = (prefix, prefix, prefix)

    command = 'genhub-stats.py --species ' + db.label
    command += ' --iloci %s.iloci.gff3 %s.iloci.fa %s.iloci.tsv' % prefix3
    command += ' --miloci %s.miloci.gff3 %s.miloci.fa %s.miloci.tsv' % prefix3
    command += (' --prnas %s.ilocus.mrnas.gff3 %s.pre-mrnas.fa '
                '%s.pre-mrnas.tsv' % prefix3)
    command += ' --mrnas %s.mrnas.gff3 %s.mrnas.fa %s.mrnas.tsv' % prefix3
    command += ' --cds %s.ilocus.mrnas.gff3 %s.cds.fa %s.cds.tsv' % prefix3
    command += (' --exons %s.ilocus.mrnas.gff3 %s.exons.fa '
                '%s.exons.tsv' % prefix3)
    command += (' --introns %s.ilocus.mrnas.gff3 %s.introns.fa '
                '%s.introns.tsv' % prefix3)

    cmd = command.split(' ')
    subprocess.check_call(cmd)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import filecmp
import re
import subprocess
import sys
import genhub


def intervals(db, delta=500, ilcformat='{}ILC-%05lu', logstream=sys.stderr):
    """
    Compute iLocus intervals.

    See the documentation for LocusPocus and the AEGeAn Toolkit for more
    details.
    """
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] computing interval loci' % db.config['species']
        print(logmsg, file=logstream)

    nameformat = ilcformat.format(db.label)
    specdir = '%s/%s' % (db.workdir, db.label)
    command = 'lpdriver.py --namefmt=%s' % nameformat
    command += ' --delta=%d' % delta
    command += ' --ilenfile=%s/ilens.temp' % specdir
    command += ' --out=%s/%s.iloci.gff3' % (specdir, db.label)
    command += ' %s/%s.gff3' % (specdir, db.label)
    cmd = command.split(' ')
    subprocess.check_call(cmd)

    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] merging iLoci' % db.config['species']
        print(logmsg, file=logstream)
    infile = '%s/%s.iloci.gff3' % (specdir, db.label)
    outfile = '%s/%s.miloci.gff3' % (specdir, db.label)
    with open(infile, 'r') as instream, open(outfile, 'w') as outstream:
        subprocess.check_call('miloci.py', stdin=instream, stdout=outstream)


def simple(db, logstream=sys.stderr):
    """Determine simple iLoci (those containing a single gene)."""
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] determining simple iLoci' % db.config['species']
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)
    ilocusfile = '%s/%s.iloci.gff3' % (specdir, db.label)
    outfile = '%s/%s.simple-iloci.txt' % (specdir, db.label)
    with open(ilocusfile, 'r') as instream, open(outfile, 'w') as outstream:
        for line in instream:
            if '\tlocus\t' not in line:
                continue
            if 'child_mRNA=' in line and 'child_gene=1;' in line:
                namematch = re.search('Name=([^;\n]+)', line)
                ilocusname = namematch.group(1)
                print(ilocusname, file=outstream)


def representatives(db, logstream=sys.stderr):
    """Select a single representative gene model for each iLocus."""
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'selecting iLocus representatives'
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)
    infile = '%s/%s.iloci.gff3' % (specdir, db.label)
    outfile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    mapfile = '%s/%s.ilocus.mrnas.txt' % (specdir, db.label)
    grepproc = subprocess.Popen(['grep', '-v', '\tintron\t', infile],
                                stdout=subprocess.PIPE,
                                universal_newlines=True)
    pmrnaproc = subprocess.Popen(['pmrna', '--locus', '--map=%s' % mapfile],
                                 stdin=grepproc.stdout,
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True)
    canonproc = subprocess.Popen(['canon-gff3', '--outfile=%s' % outfile],
                                 stdin=pmrnaproc.stdout,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 universal_newlines=True)
    stdout, stderr = canonproc.communicate()
    for line in stderr.split('\n'):
        if 'no valid mRNAs' not in line and line != '':  # pragma: no cover
            print(line, file=logstream)


def sequences(db, logstream=sys.stderr):
    """Extract iLocus sequences."""
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'extracting iLocus sequences'
        print(logmsg, file=logstream)

    specdir = '%s/%s' % (db.workdir, db.label)
    fastain = '%s/%s.gdna.fa' % (specdir, db.label)
    for ltype in ['iloci', 'miloci']:
        outfile = '%s/%s.%s.fa' % (specdir, db.label, ltype)
        gff3in = '%s/%s.%s.gff3' % (specdir, db.label, ltype)
        command = 'xtractore --type=locus '
        command += '--outfile=%s %s %s' % (outfile, gff3in, fastain)
        cmd = command.split(' ')
        proc = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                                universal_newlines=True)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               line != '':  # pragma: no cover
                print(line, file=logstream)
        assert proc.returncode == 0, 'command failed: ' + command


def ancillary(db, logstream=sys.stderr):
    """Process iLocus ancillary data."""
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] iLoci ancillary data' % db.config['species']
        print(logmsg, file=logstream)

    specdir = '%s/%s' % (db.workdir, db.label)
    cmd = ['sed', 's/^/%s\t/' % db.label, '%s/ilens.temp' % specdir]
    ilensfile = '%s/%s.ilens.tsv' % (specdir, db.label)
    with open(ilensfile, 'w') as outstream:
        subprocess.check_call(cmd, stdout=outstream)

    ilocusfile = '%s/%s.iloci.gff3' % (specdir, db.label)
    cmd = ['genhub-filens.py', db.label, ilocusfile]
    filensfile = '%s/%s.filens.tsv' % (specdir, db.label)
    with open(filensfile, 'w') as outstream:
        subprocess.check_call(cmd, stdout=outstream)

    infile = '%s/%s.ilocus.mrnas.txt' % (specdir, db.label)
    outfile = '%s/%s.mrnas.txt' % (specdir, db.label)
    with open(outfile, 'w') as outstream:
        subprocess.check_call(['cut', '-f', '2', infile], stdout=outstream)


# -----------------------------------------------------------------------------
# Driver function
# -----------------------------------------------------------------------------

def prepare(db, delta=500, ilcformat='%sILC-', logstream=sys.stderr):  # pragma: no cover # noqa
    intervals(db, delta=delta, ilcformat=ilcformat, logstream=logstream)
    simple(db, logstream=logstream)
    representatives(db, logstream=logstream)
    sequences(db, logstream=logstream)
    ancillary(db, logstream=logstream)


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_intervals():
    """Parse iLocus intervals"""
    config = genhub.test_registry.genome('Bdis')
    d = genhub.refseq.RefSeqDB('Bdis', config, workdir='testdata/demo-workdir')
    intervals(d, logstream=None)

    outfile = 'testdata/demo-workdir/Bdis/Bdis.iloci.gff3'
    testfile = 'testdata/gff3/bdis-iloci.gff3'
    assert filecmp.cmp(outfile, testfile), 'iLocus parsing failed'

    outfile = 'testdata/demo-workdir/Bdis/Bdis.miloci.gff3'
    testfile = 'testdata/gff3/bdis-miloci.gff3'
    assert filecmp.cmp(outfile, testfile), 'miLocus parsing failed'


def test_simple():
    """Determine simple iLoci"""
    config = genhub.test_registry.genome('Bdis')
    d = genhub.refseq.RefSeqDB('Bdis', config, workdir='testdata/demo-workdir')
    simple(d, logstream=None)

    outfile = 'testdata/demo-workdir/Bdis/Bdis.simple-iloci.txt'
    testfile = 'testdata/misc/bdis-simple.txt'
    assert filecmp.cmp(outfile, testfile), 'simple iLocus ID failed'


def test_reps():
    """Select representative gene models for each iLocus"""
    config = genhub.test_registry.genome('Bdis')
    d = genhub.refseq.RefSeqDB('Bdis', config, workdir='testdata/demo-workdir')
    representatives(d, logstream=None)

    outfile = 'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.gff3'
    testfile = 'testdata/gff3/bdis-reps.gff3'
    assert filecmp.cmp(outfile, testfile), 'iLocus rep ID failed'


def test_sequences():
    """Extract iLocus sequences"""
    config = genhub.test_registry.genome('Bdis')
    d = genhub.refseq.RefSeqDB('Bdis', config, workdir='testdata/demo-workdir')
    sequences(d, logstream=None)

    outfile = 'testdata/demo-workdir/Bdis/Bdis.iloci.fa'
    testfile = 'testdata/fasta/bdis-iloci.fa'
    assert filecmp.cmp(outfile, testfile), 'iLocus sequence extraction failed'

    outfile = 'testdata/demo-workdir/Bdis/Bdis.miloci.fa'
    testfile = 'testdata/fasta/bdis-miloci.fa'
    assert filecmp.cmp(outfile, testfile), 'miLocus sequence extraction failed'


def test_ancillary():
    """Process ancillary data for iLoci"""
    config = genhub.test_registry.genome('Bdis')
    d = genhub.refseq.RefSeqDB('Bdis', config, workdir='testdata/demo-workdir')
    ancillary(d, logstream=None)

    outfile = 'testdata/demo-workdir/Bdis/Bdis.ilens.tsv'
    testfile = 'testdata/misc/bdis-ilens.tsv'
    assert filecmp.cmp(outfile, testfile), 'iLocus length record failed'

    outfile = 'testdata/demo-workdir/Bdis/Bdis.filens.tsv'
    testfile = 'testdata/misc/bdis-filens.tsv'
    assert filecmp.cmp(outfile, testfile), 'flanking iLocus length failed'

    outfile = 'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.txt'
    testfile = 'testdata/misc/bdis-ilocus-mrnas.txt'
    assert filecmp.cmp(outfile, testfile), 'iLocus rep cut failed'



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Module implementing a registry for handling genome configuration files."""

from __future__ import print_function
import glob
import os
import pkg_resources
import yaml
import genhub
try:
    FileNotFoundError
except NameError:  # pragma: no cover
    FileNotFoundError = IOError


class Registry(object):

    def __init__(self):
        """Initialize the registry with the default GenHub configs."""
        genhubdir = pkg_resources.resource_filename('genhub', 'genomes')
        self.update(genhubdir, clear=True)

    def update(self, path, clear=False):
        """
        Update the registry from the given directory path.

        The registry will attempt to load all .yml files as genome configs and
        .txt files as batch configs. If `clear` is true, any previous entries
        in the registry will be cleared before loading new entries.
        """
        if clear:
            self.genome_configs = dict()
            self.batch_configs = dict()

        if not os.path.exists(path):
            message = 'config directory "%s" does not exist' % path
            raise FileNotFoundError(message)

        for filepath in glob.glob(path + '/*.yml'):
            configs = self.parse_genome_config(filepath)
            self.genome_configs.update(configs)

        for filepath in glob.glob(path + '/*.txt'):
            batch = self.parse_batch_config(filepath)
            filename = os.path.basename(filepath)
            batch_label = os.path.splitext(filename)[0]
            self.batch_configs[batch_label] = batch

    def genome(self, label):
        """Retrieve a genome config from the registry by label."""
        if label not in self.genome_configs:
            return None
        return self.genome_configs[label]

    def genomes(self, labels):
        config = dict()
        for label in labels:
            config[label] = self.genome_configs[label]
        return config

    def batch(self, batch_label):
        """Retrieve a batch of genome configs from the registry."""
        if batch_label not in self.batch_configs:
            return None
        config = dict()
        for genome_label in self.batch_configs[batch_label]:
            config[genome_label] = self.genome_configs[genome_label]
        return config

    def parse_genome_config(self, config):
        """
        Parse genome config in YAML format.

        If `config` is a string it is treated as a filename, otherwise as a
        file handle or similar object.
        """
        if isinstance(config, str):
            config = open(config, 'r')
        return yaml.load(config)

    def parse_batch_config(self, config):
        """
        Parse a batch of genome config labels from a file.

        The file should contain a single genome config label per line, and
        `config` is treated as a filename.
        """
        batch = list()
        with open(config, 'r') as infile:
            for line in infile:
                label = line.strip()
                if label != '':
                    batch.append(label)
        return batch

    def list_genomes(self):
        for label in sorted(self.genome_configs):
            yield label, self.genome_configs[label]

    def list_batches(self):
        for label in sorted(self.batch_configs):
            yield label, self.batch_configs[label]


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_list():
    """Listing genome and batch configs"""
    registry = Registry()

    genome_labels = [x for x in registry.list_genomes()]
    ymlfiles = [x for x in glob.glob('genhub/genomes/*.yml')]
    assert len(genome_labels) == len(ymlfiles)
    batch_labels = [x for x in registry.list_batches()]
    txtfiles = [x for x in glob.glob('genhub/genomes/*.txt')]
    assert len(batch_labels) == len(txtfiles)

    registry.update('testdata/conf', clear=True)

    genome_labels = [x for x in registry.list_genomes()]
    ymlfiles = [x for x in glob.glob('testdata/conf/*.yml')]
    assert len(genome_labels) == len(ymlfiles)
    batch_labels = [x for x in registry.list_batches()]
    txtfiles = [x for x in glob.glob('testdata/conf/*.txt')]
    assert len(batch_labels) == len(txtfiles)

    try:
        registry.update('foobar/bogus')
    except FileNotFoundError:
        pass


def test_genome():
    """Loading a genome configuration by label"""
    registry = Registry()
    config = registry.genome('Osat')
    assert 'accession' in config
    assert config['accession'] == 'GCF_000005425.2'

    registry.update('testdata/conf')
    config = registry.genome('Osat')
    assert 'accession' in config
    assert config['accession'] == 'BogusThisIsNotARealAccession'

    config = registry.genome('Bvul')
    assert 'scaffolds' in config
    assert config['scaffolds'] == 'bv_ref_1.1_chrUn.fa.gz'

    config = registry.genomes(['Atha', 'Bdis', 'Osat'])
    assert len(config) == 3
    assert sorted(config) == ['Atha', 'Bdis', 'Osat']

    assert registry.genome('Docc') is None


def test_batch():
    """Loading a batch configuration by label"""
    registry = Registry()
    config = registry.batch('honeybees')
    assert len(config) == 3
    assert sorted(config) == ['Ador', 'Aflo', 'Amel']

    registry.update('testdata/conf')
    config = registry.batch('mythical')
    assert len(config) == 1
    assert sorted(config) == ['Bvul']

    assert registry.batch('nonexistent') is None

    config = registry.batch('bumblebees')
    assert len(config) == 2
    assert sorted(config) == ['Bimp', 'Bter']
    assert config['Bter']['accession'] == 'GCF_000214255.1'
    assert config['Bimp']['accession'] == 'BogusAccessionForUnitTesting'


def test_parse_genome_config():
    """Parsing genome configurations from a file"""
    registry = Registry()
    with open('genhub/genomes/Pbar.yml', 'r') as filehandle:
        config = registry.parse_genome_config(filehandle)
        assert len(config) == 1
        assert 'Pbar' in config
        assert config['Pbar']['species'] == 'Pogonomyrmex barbatus'

    config = registry.parse_genome_config('genhub/genomes/Hlab.yml')
    assert len(config) == 1
    assert 'Hlab' in config
    assert config['Hlab']['common'] == 'blueberry bee'
